@@ -2,9 +2,45 @@
 #define MARPAESLIF_H
 
 #include <stddef.h>                 /* For size_t */
+#include <stdio.h>                  /* For __int64, eventually */
+#include <limits.h>                 /* For long long, eventually */
 #include <genericLogger.h>
 #include <marpaESLIF/export.h>
 
+#if defined(LLONG_MAX) && defined(LLONG_MIN)
+#  define MARPAESLIF_HAVE_LONG_LONG
+#  define MARPAESLIF_LONG_LONG long long
+#  define MARPAESLIF_LLONG_MAX LLONG_MAX
+#  define MARPAESLIF_LLONG_MIN LLONG_MIN
+#  define MARPAESLIF_LONG_LONG_FMT "%lld"
+#else
+#  if defined(LONGLONG_MAX) && defined(LONGLONG_MIN)
+#    define MARPAESLIF_HAVE_LONG_LONG
+#    define MARPAESLIF_LONG_LONG long long
+#    define MARPAESLIF_LLONG_MAX LONGLONG_MAX
+#    define MARPAESLIF_LLONG_MIN LONGLONG_MIN
+#    define MARPAESLIF_LONG_LONG_FMT "%lld"
+#  else
+#    if defined(_MSC_VER) || defined(__BORLANDC__)
+/*     Just because of versions of these compilers might not have long long, but they */
+/*     always had __int64. Note that on Windows short is always 2, int is always 4,   */
+/*     long is always 4, __int64 is always 8 */
+#      define MARPAESLIF_HAVE_LONG_LONG
+#      define MARPAESLIF_LONG_LONG __int64
+#      if defined(_I64_MAX) && defined(_I64_MIN)
+#        define MARPAESLIF_LLONG_MAX _I64_MAX
+#        define MARPAESLIF_LLONG_MIN _I64_MIN
+#      else
+#        define MARPAESLIF_LLONG_MAX 9223372036854775807i64
+#        define MARPAESLIF_LLONG_MIN -9223372036854775808i64
+#      endif
+#      define MARPAESLIF_LONG_LONG_FMT "%I64d"
+#    else
+#      undef MARPAESLIF_HAVE_LONG_LONG
+#    endif
+#  endif
+#endif
+
 typedef struct marpaESLIFOption {
   genericLogger_t *genericLoggerp;  /* Logger. Default: NULL */
 } marpaESLIFOption_t;
@@ -16,6 +52,9 @@ typedef struct marpaESLIFValue       marpaESLIFValue_t;
 typedef struct marpaESLIFSymbol      marpaESLIFSymbol_t;
 typedef struct marpaESLIFValueResult marpaESLIFValueResult_t;
 
+/* General free callback */
+typedef void  (*marpaESLIFFreeCallback_t)(void *userDatavp, marpaESLIFValueResult_t *marpaESLIFValueResultp);
+
 /* A string */
 typedef struct marpaESLIFString {
   char   *bytep;            /* pointer bytes */
@@ -33,21 +72,22 @@ typedef struct marpaESLIFGrammarOption {
   size_t  bytel;               /* Input length in byte unit */
   char   *encodings;           /* Input encoding. Default: NULL */
   size_t  encodingl;           /* Length of encoding itself. Default: 0 */
-  char   *encodingOfEncodings; /* Encoding of encoding, in ASCII encoding. Default: NULL. */
 } marpaESLIFGrammarOption_t;
 
-/* The reader can return encoding information, giving eventual encoding of this information in encodingOfEncodingsp, starting at *encodingsp, spreaded over *encodinglp bytes */
-typedef short (*marpaESLIFReader_t)(void *userDatavp, char **inputcpp, size_t *inputlp, short *eofbp, short *characterStreambp, char **encodingOfEncodingsp, char **encodingsp, size_t *encodinglp);
+
+/* The reader can return encoding information, giving eventual encoding in *encodingsp, spreaded over *encodinglp bytes. Encoding of encoding is free. */
+typedef short (*marpaESLIFReader_t)(void *userDatavp, char **inputcpp, size_t *inputlp, short *eofbp, short *characterStreambp, char **encodingsp, size_t *encodinglp);
+
 typedef struct marpaESLIFRecognizerOption {
-  void                *userDatavp;                  /* User specific context */
-  marpaESLIFReader_t   marpaESLIFReaderCallbackp;   /* Reader */
-  short                disableThresholdb;           /* Default: 0 */
-  short                exhaustedb;                  /* Exhaustion event. Default: 0 */
-  short                newlineb;                    /* Count line/column numbers. Default: 0 */
-  short                trackb;                      /* Track absolute position. Default: 0 */
-  size_t               bufsizl;                     /* Minimum stream buffer size: Recommended: 0 (internally, a system default will apply) */
-  unsigned int         buftriggerperci;             /* Excess number of bytes, in percentage of bufsizl, where stream buffer size is reduced. Recommended: 50 */
-  unsigned int         bufaddperci;                 /* Policy of minimum of bytes for increase, in percentage of current allocated size, when stream buffer size need to augment. Recommended: 50 */
+  void               *userDatavp;          /* User specific context */
+  marpaESLIFReader_t  readerCallbackp;     /* Reader */
+  short               disableThresholdb;   /* Default: 0 */
+  short               exhaustedb;          /* Exhaustion event. Default: 0 */
+  short               newlineb;            /* Count line/column numbers. Default: 0 */
+  short               trackb;              /* Track absolute position. Default: 0 */
+  size_t              bufsizl;             /* Minimum stream buffer size: Recommended: 0 (internally, a system default will apply) */
+  unsigned int        buftriggerperci;     /* Excess number of bytes, in percentage of bufsizl, where stream buffer size is reduced. Recommended: 50 */
+  unsigned int        bufaddperci;         /* Policy of minimum of bytes for increase, in percentage of current allocated size, when stream buffer size need to augment. Recommended: 50 */
 } marpaESLIFRecognizerOption_t;
 
 typedef enum marpaESLIFEventType {
@@ -67,14 +107,6 @@ typedef struct marpaESLIFEvent {
   char                 *events;  /* Event name, always NULL if exhaustion eent */
 } marpaESLIFEvent_t;
 
-typedef short (*marpaESLIFValueRuleCallback_t)(void *userDatavp, marpaESLIFValue_t *marpaESLIFValuep, int arg0i, int argni, int resulti, short nullableb);
-typedef short (*marpaESLIFValueSymbolCallback_t)(void *userDatavp, marpaESLIFValue_t *marpaESLIFValuep, char *bytep, size_t bytel, int resulti);
-typedef void  (*marpaESLIFValueFreeCallback_t)(void *userDatavp, int contexti, void *p, size_t sizel);
-
-typedef marpaESLIFValueRuleCallback_t   (*marpaESLIFValueRuleActionResolver_t)(void *userDatavp, marpaESLIFValue_t *marpaESLIFValuep, char *actions);
-typedef marpaESLIFValueSymbolCallback_t (*marpaESLIFValueSymbolActionResolver_t)(void *userDatavp, marpaESLIFValue_t *marpaESLIFValuep, char *actions);
-typedef marpaESLIFValueFreeCallback_t   (*marpaESLIFValueFreeActionResolver_t)(void *userDatavp, marpaESLIFValue_t *marpaESLIFValuep, char *actions);
-
 /* Value types */
 typedef enum marpaESLIFValueType {
   MARPAESLIF_VALUE_TYPE_UNDEF = 0,
@@ -85,29 +117,110 @@ typedef enum marpaESLIFValueType {
   MARPAESLIF_VALUE_TYPE_FLOAT,
   MARPAESLIF_VALUE_TYPE_DOUBLE,
   MARPAESLIF_VALUE_TYPE_PTR,
-  MARPAESLIF_VALUE_TYPE_ARRAY
+  MARPAESLIF_VALUE_TYPE_ARRAY,
+  MARPAESLIF_VALUE_TYPE_BOOL,
+  MARPAESLIF_VALUE_TYPE_STRING,
+  MARPAESLIF_VALUE_TYPE_ROW,
+  MARPAESLIF_VALUE_TYPE_TABLE,
+  MARPAESLIF_VALUE_TYPE_LONG_DOUBLE
+#ifdef MARPAESLIF_HAVE_LONG_LONG
+  ,MARPAESLIF_VALUE_TYPE_LONG_LONG
+#endif
 } marpaESLIFValueType_t;
 
+/* Callback definitions */
+typedef short (*marpaESLIFValueRuleCallback_t)(void *userDatavp, marpaESLIFValue_t *marpaESLIFValuep, int arg0i, int argni, int resulti, short nullableb);
+typedef short (*marpaESLIFValueSymbolCallback_t)(void *userDatavp, marpaESLIFValue_t *marpaESLIFValuep, marpaESLIFValueResult_t *marpaESLIFValueResult, int resulti);
+
+typedef marpaESLIFValueRuleCallback_t   (*marpaESLIFValueRuleActionResolver_t)(void *userDatavp, marpaESLIFValue_t *marpaESLIFValuep, char *actions);
+typedef marpaESLIFValueSymbolCallback_t (*marpaESLIFValueSymbolActionResolver_t)(void *userDatavp, marpaESLIFValue_t *marpaESLIFValuep, char *actions);
+
 /* Valuation result */
-/* The representation returns a sequence of bytes and is appended AS-IS */
+/* The representation returns a sequence of bytes, eventually meaning a string */
 /* It is legal to return NULL in *inputcpp or 0 in *inputlp: representation will be ignored */
-typedef short (*marpaESLIFRepresentation_t)(void *userDatavp, marpaESLIFValueResult_t *marpaESLIFValueResultp, char **inputcpp, size_t *inputlp);
+/* It is legal to return NULL in encodingmaybesp, then this is an opaque sequence of bytes, else */
+/* this is considered as a string in this given encoding. */
+/* Note that it is the responsibility of the caller to make sure that *inputcpp and **encodingmaybesp points to valid memory area when the call returns */
+typedef short (*marpaESLIFRepresentation_t)(void *userDatavp, marpaESLIFValueResult_t *marpaESLIFValueResultp, char **inputcpp, size_t *inputlp, char **encodingasciisp);
+typedef char marpaESLIFValueResultChar_t;
+typedef short marpaESLIFValueResultShort_t;
+typedef int marpaESLIFValueResultInt_t;
+typedef long marpaESLIFValueResultLong_t;
+typedef float marpaESLIFValueResultFloat_t;
+typedef double marpaESLIFValueResultDouble_t;
+typedef struct marpaESLIFValueResultPtr {
+  void                     *p;
+  void                     *freeUserDatavp;
+  marpaESLIFFreeCallback_t  freeCallbackp;
+  short                     shallowb;
+} marpaESLIFValueResultPtr_t;
+typedef struct marpaESLIFValueResultArray {
+  char                     *p;
+  void                     *freeUserDatavp;
+  marpaESLIFFreeCallback_t  freeCallbackp;
+  short                     shallowb;
+  size_t                    sizel;
+} marpaESLIFValueResultArray_t;
+typedef enum marpaESLIFValueResultBool {
+  MARPAESLIFVALUERESULTBOOL_FALSE = 0,
+  MARPAESLIFVALUERESULTBOOL_TRUE = 1
+} marpaESLIFValueResultBool_t;
+typedef struct marpaESLIFValueResultString {
+  unsigned char            *p;
+  void                     *freeUserDatavp;
+  marpaESLIFFreeCallback_t  freeCallbackp;
+  short                     shallowb;
+  size_t                    sizel;
+  char                     *encodingasciis;
+} marpaESLIFValueResultString_t;
+typedef struct marpaESLIFValueResultRow {
+  marpaESLIFValueResult_t  *p;
+  void                     *freeUserDatavp;
+  marpaESLIFFreeCallback_t  freeCallbackp;
+  short                     shallowb;
+  size_t                    sizel;
+} marpaESLIFValueResultRow_t;
+typedef struct marpaESLIFValueResultPair marpaESLIFValueResultPair_t;
+typedef struct marpaESLIFValueResultTable {
+  marpaESLIFValueResultPair_t *p;
+  void                        *freeUserDatavp;
+  marpaESLIFFreeCallback_t     freeCallbackp;
+  short                        shallowb;
+  size_t                       sizel;
+} marpaESLIFValueResultTable_t;
+typedef long double marpaESLIFValueResultLongDouble_t;
+#ifdef MARPAESLIF_HAVE_LONG_LONG
+typedef MARPAESLIF_LONG_LONG marpaESLIFValueResultLongLong_t;
+#endif
+
 struct marpaESLIFValueResult {
-  int                        contexti;          /* Free value meaningful only to the user */
-  size_t                     sizel;             /* Length of data in case value is an ARRAY - always 0 otherwise */
+  void                      *contextp;          /* Free value meaningful only to the user */
   marpaESLIFRepresentation_t representationp;   /* How a user-land alternative is represented if it was in the input */
-  short                      shallowb;          /* In case of PTR or ARRAY, indicate if this is a shallow pointer */
   marpaESLIFValueType_t      type;              /* Type for tagging the following union */
   union {
-    char    c;                                  /* Value is a char */
-    short   b;                                  /* Value is a short */
-    int     i;                                  /* Value is an int */
-    long    l;                                  /* Value is a long */
-    float   f;                                  /* Value is a float */
-    double  d;                                  /* Value is a double */
-    void   *p;                                  /* Value is a pointer */
+    marpaESLIFValueResultChar_t        c; /* Value is a char */
+    marpaESLIFValueResultShort_t       b; /* Value is a short */
+    marpaESLIFValueResultInt_t         i; /* Value is an int */
+    marpaESLIFValueResultLong_t        l; /* Value is a long */
+    marpaESLIFValueResultFloat_t       f; /* Value is a float */
+    marpaESLIFValueResultDouble_t      d; /* Value is a double */
+    marpaESLIFValueResultPtr_t         p; /* Value is a pointer */
+    marpaESLIFValueResultArray_t       a; /* Value is a byte array */
+    marpaESLIFValueResultBool_t        y; /* Value is a boolean */
+    marpaESLIFValueResultString_t      s; /* Value is a string */
+    marpaESLIFValueResultRow_t         r; /* Value is a row of values */
+    marpaESLIFValueResultTable_t       t; /* Value is a row of values, where sizel is even */
+    marpaESLIFValueResultLongDouble_t ld; /* Value is a long double */
+#ifdef MARPAESLIF_HAVE_LONG_LONG
+    marpaESLIFValueResultLongLong_t   ll; /* Value is a long long */
+#endif
   } u;
 };
+/* Now that marpaESLIFValueResult is defined, we can define the pair */
+struct marpaESLIFValueResultPair {
+  marpaESLIFValueResult_t    key;
+  marpaESLIFValueResult_t    value;
+};
 
 /* An alternative from external lexer point of view */
 typedef struct marpaESLIFAlternative {
@@ -116,11 +229,14 @@ typedef struct marpaESLIFAlternative {
   size_t                     grammarLengthl; /* Length within the grammar (1 in the token-stream model) */
 } marpaESLIFAlternative_t;
 
+/* Ask the host system to import a marpaESLIFValueResult */
+typedef short (*marpaESLIFValueResultImport_t)(marpaESLIFValue_t *marpaESLIFValuep, void *userDatavp, marpaESLIFValueResult_t *marpaESLIFValueResultp);
+
 typedef struct marpaESLIFValueOption {
   void                                 *userDatavp;            /* User specific context */
   marpaESLIFValueRuleActionResolver_t   ruleActionResolverp;   /* Will return the function doing the wanted rule action */
   marpaESLIFValueSymbolActionResolver_t symbolActionResolverp; /* Will return the function doing the wanted symbol action */
-  marpaESLIFValueFreeActionResolver_t   freeActionResolverp;   /* Will return the function doing the free */
+  marpaESLIFValueResultImport_t         importerp;             /* Will ask end-user to import a marpaESLIFValueResult */
   short                                 highRankOnlyb;         /* Default: 1 */
   short                                 orderByRankb;          /* Default: 1 */
   short                                 ambiguousb;            /* Default: 0 */
@@ -137,7 +253,8 @@ typedef struct marpaESLIFRecognizerProgress {
 
 typedef enum marpaESLIFActionType {
   MARPAESLIF_ACTION_TYPE_NAME = 0,
-  MARPAESLIF_ACTION_TYPE_STRING
+  MARPAESLIF_ACTION_TYPE_STRING,
+  MARPAESLIF_ACTION_TYPE_LUA
 } marpaESLIFActionType_t;
 
 typedef struct marpaESLIFAction {
@@ -145,13 +262,13 @@ typedef struct marpaESLIFAction {
   union {
     char               *names;
     marpaESLIFString_t *stringp;
+    char               *luas;
   } u;
 } marpaESLIFAction_t;
 
 typedef struct marpaESLIFGrammarDefaults {
   marpaESLIFAction_t *defaultRuleActionp;   /* Default action for rules */
   marpaESLIFAction_t *defaultSymbolActionp; /* Default action for symbols */
-  marpaESLIFAction_t *defaultFreeActionp;   /* Default action for free - if not NULL, type can never be MARPAESLIF_ACION_TYPE_STRING */
 } marpaESLIFGrammarDefaults_t;
 
 /* Rule property */
@@ -181,7 +298,6 @@ typedef struct marpaESLIFGrammarProperty {
   short                  latmb;                        /* LATM ? */
   marpaESLIFAction_t    *defaultSymbolActionp;         /* Default action for symbols - never NULL */
   marpaESLIFAction_t    *defaultRuleActionp;           /* Default action for rules - never NULL */
-  marpaESLIFAction_t    *defaultFreeActionp;           /* Default action for free - can be NULL */
   int                    starti;                       /* Start symbol Id - always >= 0 */
   int                    discardi;                     /* Discard symbol Id (-1 if none) */
   size_t                 nsymboll;                     /* Number of symbols - always > 0*/
@@ -198,6 +314,7 @@ typedef struct marpaESLIFRuleProperty {
   int                    separatori;                   /* Eventual separator symbol Id (-1 if none) */
   size_t                 nrhsl;                        /* Number of RHS, 0 in case of a nullable */
   int                   *rhsip;                        /* Array of RHS Ids, NULL in case of a nullable */
+  short                 *skipbp;                       /* Skip flags per RHS, first RHS is at indice 0 and so on, NULL in case of a nullable or a sequence */
   int                    exceptioni;                   /* Exception symbol Id (-1 if none) */
   marpaESLIFAction_t    *actionp;                      /* Action */
   char                  *discardEvents;                /* Discard event name - shallowed with its RHS */
@@ -247,8 +364,11 @@ typedef struct marpaESLIFSymbolProperty {
 #ifdef __cplusplus
 extern "C" {
 #endif
-  marpaESLIF_EXPORT const char                   *marpaESLIF_versions(void);
   marpaESLIF_EXPORT marpaESLIF_t                 *marpaESLIF_newp(marpaESLIFOption_t *marpaESLIFOptionp);
+  marpaESLIF_EXPORT short                         marpaESLIF_versionb(marpaESLIF_t *marpaESLIFp, char **versionsp);
+  marpaESLIF_EXPORT short                         marpaESLIF_versionMajorb(marpaESLIF_t *marpaESLIFp, int *versionMajorip);
+  marpaESLIF_EXPORT short                         marpaESLIF_versionMinorb(marpaESLIF_t *marpaESLIFp, int *versionMinorip);
+  marpaESLIF_EXPORT short                         marpaESLIF_versionPatchb(marpaESLIF_t *marpaESLIFp, int *versionPatchip);
   marpaESLIF_EXPORT short                         marpaESLIF_extend_builtin_actionb(marpaESLIF_t *marpaESLIFp, char **actionsArrayp, size_t actionsArrayl);
   marpaESLIF_EXPORT marpaESLIFOption_t           *marpaESLIF_optionp(marpaESLIF_t *marpaESLIFp);
   marpaESLIF_EXPORT marpaESLIFGrammar_t          *marpaESLIF_grammarp(marpaESLIF_t *marpaESLIFp);
@@ -282,8 +402,9 @@ extern "C" {
   marpaESLIF_EXPORT short                         marpaESLIFGrammar_ruleshowform_by_levelb(marpaESLIFGrammar_t *marpaESLIFGrammarp, int rulei, char **ruleshowsp, int leveli, marpaESLIFString_t *descp);
   marpaESLIF_EXPORT short                         marpaESLIFGrammar_grammarshowform_currentb(marpaESLIFGrammar_t *marpaESLIFGrammarp, char **grammarshowsp);
   marpaESLIF_EXPORT short                         marpaESLIFGrammar_grammarshowform_by_levelb(marpaESLIFGrammar_t *marpaESLIFGrammarp, char **grammarshowsp, int leveli, marpaESLIFString_t *descp);
-  marpaESLIF_EXPORT short                         marpaESLIFGrammar_parseb(marpaESLIFGrammar_t *marpaESLIFGrammarp, marpaESLIFRecognizerOption_t *marpaESLIFRecognizerOptionp, marpaESLIFValueOption_t *marpaESLIFValueOptionp, short *exhaustedbp, marpaESLIFValueResult_t *marpaESLIFValueResultp);
-  marpaESLIF_EXPORT short                         marpaESLIFGrammar_parse_by_levelb(marpaESLIFGrammar_t *marpaESLIFGrammarp, marpaESLIFRecognizerOption_t *marpaESLIFRecognizerOptionp, marpaESLIFValueOption_t *marpaESLIFValueOptionp, short *exhaustedbp, int leveli, marpaESLIFString_t *descp, marpaESLIFValueResult_t *marpaESLIFValueResultp);
+  marpaESLIF_EXPORT short                         marpaESLIFGrammar_grammarshowscriptb(marpaESLIFGrammar_t *marpaESLIFGrammarp, char **grammarscriptsp);
+  marpaESLIF_EXPORT short                         marpaESLIFGrammar_parseb(marpaESLIFGrammar_t *marpaESLIFGrammarp, marpaESLIFRecognizerOption_t *marpaESLIFRecognizerOptionp, marpaESLIFValueOption_t *marpaESLIFValueOptionp, short *exhaustedbp);
+  marpaESLIF_EXPORT short                         marpaESLIFGrammar_parse_by_levelb(marpaESLIFGrammar_t *marpaESLIFGrammarp, marpaESLIFRecognizerOption_t *marpaESLIFRecognizerOptionp, marpaESLIFValueOption_t *marpaESLIFValueOptionp, short *exhaustedbp, int leveli, marpaESLIFString_t *descp);
   marpaESLIF_EXPORT void                          marpaESLIFGrammar_freev(marpaESLIFGrammar_t *marpaESLIFGrammarp);
 
   marpaESLIF_EXPORT marpaESLIFRecognizer_t       *marpaESLIFRecognizer_newp(marpaESLIFGrammar_t *marpaESLIFGrammarp, marpaESLIFRecognizerOption_t *marpaESLIFRecognizerOptionp);
@@ -319,23 +440,52 @@ extern "C" {
   marpaESLIF_EXPORT marpaESLIFValue_t            *marpaESLIFValue_newp(marpaESLIFRecognizer_t *marpaESLIFRecognizerp, marpaESLIFValueOption_t *marpaESLIFValueOptionp);
   marpaESLIF_EXPORT marpaESLIFRecognizer_t       *marpaESLIFValue_recognizerp(marpaESLIFValue_t *marpaESLIFValuep);
   marpaESLIF_EXPORT marpaESLIFValueOption_t      *marpaESLIFValue_optionp(marpaESLIFValue_t *marpaESLIFValuep);
-  marpaESLIF_EXPORT short                         marpaESLIFValue_valueb(marpaESLIFValue_t *marpaESLIFValuep, marpaESLIFValueResult_t *marpaESLIFValueResultp);
+  marpaESLIF_EXPORT short                         marpaESLIFValue_valueb(marpaESLIFValue_t *marpaESLIFValuep);
   marpaESLIF_EXPORT short                         marpaESLIFValue_value_startb(marpaESLIFValue_t *marpaESLIFValuep, int *startip);
   marpaESLIF_EXPORT short                         marpaESLIFValue_value_lengthb(marpaESLIFValue_t *marpaESLIFValuep, int *lengthip);
   marpaESLIF_EXPORT short                         marpaESLIFValue_contextb(marpaESLIFValue_t *marpaESLIFValuep, char **symbolsp, int *symbolip, char **rulesp, int *ruleip);
   marpaESLIF_EXPORT void                          marpaESLIFValue_freev(marpaESLIFValue_t *marpaESLIFValuep);
 
+  /* ------------------------------------- */
   /* Stack management when doing valuation */
+  /* ------------------------------------- */
+  /* Generic setter */
   marpaESLIF_EXPORT short                         marpaESLIFValue_stack_setb(marpaESLIFValue_t *marpaESLIFValuep, int indicei, marpaESLIFValueResult_t *marpaESLIFValueResultp);
-  /* forget is like setting a value of type MARPAESLIFVALUE_TYPE_UNDEF except that memory management is switched off */
-  marpaESLIF_EXPORT short                         marpaESLIFValue_stack_forgetb(marpaESLIFValue_t *marpaESLIFValuep, int indicei);
+  /* Generic getter */
   marpaESLIF_EXPORT short                         marpaESLIFValue_stack_getb(marpaESLIFValue_t *marpaESLIFValuep, int indicei, marpaESLIFValueResult_t *marpaESLIFValueResultp);
-  marpaESLIF_EXPORT marpaESLIFValueResult_t       *marpaESLIFValue_stack_getp(marpaESLIFValue_t *marpaESLIFValuep, int indicei);
+  /* Optimized generic getter: returns a direct pointer to the marpaESLIFValuep in memory instead of a copy via marpaESLIFValue_stack_getb */
+  marpaESLIF_EXPORT marpaESLIFValueResult_t      *marpaESLIFValue_stack_getp(marpaESLIFValue_t *marpaESLIFValuep, int indicei);
 
-  /* getAndForgetb transfers the memory management from the stack to the end-user in one call */
+  /* Helpers */
+  
+  /* marpaESLIFValue_stack_forgetb is like setting a value of type MARPAESLIFVALUE_TYPE_UNDEF saying that memory management is switched off at this specific indice */
+  marpaESLIF_EXPORT short                         marpaESLIFValue_stack_forgetb(marpaESLIFValue_t *marpaESLIFValuep, int indicei);
+  /* marpaESLIFValue_stack_getAndForgetb transfers the memory management from the stack to the end-user in one call */
+  /* It is nothing else but a wrapper on marpaESLIFValue_stack_getb followed by marpaESLIFValue_stack_forgetb */
   marpaESLIF_EXPORT short                         marpaESLIFValue_stack_getAndForgetb(marpaESLIFValue_t *marpaESLIFValuep, int indicei, marpaESLIFValueResult_t *marpaESLIFValueResultp);
+  /* marpaESLIFValue_importb call the end-user importerp() function callback. */
+  /* -------------------------------------------------------------------------------------------------------------------------------------------------------------- */
+  /* ROW and TABLE types are flat'idified:                                                                                                                          */
+  /* - For an ROW type, all members are imported first in order, then the row marpaESLIFValueResult itself (which contains the number of elements in u.r.sizel)     */
+  /*   i.e. value[0], value[1], ... value[value.u.r.sizel - 1], value                                                                                               */
+  /* - For an TABLE type, all members are imported first in order, then the array marpaESLIFValueResult itself (which contains the number of elements in u.t.sizel) */
+  /*   i.e. table[0], table[1], ... table[table.u.t.sizel - 1], table                                                                                               */
+  /*   Since a table member is nothign else but a row with an even number of elements, it means this can be reinterpreted as:                                       */
+  /*        key[0],value[0]...,key[(table.u.t.sizel/2)-1],value[(table.u.t.sizel/2)-1]                                                                              */
+  /*                                                                                                                                                                */
+  /*   This mean that the importer callback should maintain an internal stack of marpaESLIFValueResult_t items everytime it is called and:                          */
+  /*   - push to its stack anything that is not a ROW or a TABLE                                                                                                    */
+  /*   - pop from its stack value.u.r.sizel   elements when it receives a ROW callback                                                                              */
+  /*   - pop from its stack value.u.t.sizel*2 elements when it receives a TABLE callback                                                                            */
+  /*   At the end it must have exactly one element in its internal stack.                                                                                           */
+  /* -------------------------------------------------------------------------------------------------------------------------------------------------------------- */
+  marpaESLIF_EXPORT short                         marpaESLIFValue_importb(marpaESLIFValue_t *marpaESLIFValuep, marpaESLIFValueResult_t *marpaESLIFValueResultp, marpaESLIFValueResult_t *marpaESLIFValueResultResolvedp);
+
+  /* Helper routine to guess the encoding of a byte array. If it returns a non-NULL value it is the responsibility of the caller to free it */
+  marpaESLIF_EXPORT char                         *marpaESLIF_encodings(marpaESLIF_t *marpaESLIFp, char *bytep, size_t bytel);
 
   marpaESLIF_EXPORT void                          marpaESLIF_freev(marpaESLIF_t *marpaESLIFp);
+
 #ifdef __cplusplus
 }
 #endif
